package it.openutils.magnoliastripes;

import info.magnolia.cms.beans.config.Paragraph;
import info.magnolia.cms.beans.runtime.ParagraphRenderer;
import info.magnolia.cms.core.Content;
import info.magnolia.cms.core.NodeData;
import info.magnolia.context.MgnlContext;
import info.magnolia.context.WebContext;

import java.io.IOException;
import java.io.Writer;
import java.lang.reflect.InvocationTargetException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Stack;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.jsp.PageContext;

import net.sourceforge.stripes.action.ActionBeanContext;
import net.sourceforge.stripes.action.Resolution;
import net.sourceforge.stripes.config.Configuration;
import net.sourceforge.stripes.controller.DispatcherHelper;
import net.sourceforge.stripes.controller.DispatcherServlet;
import net.sourceforge.stripes.controller.ExecutionContext;
import net.sourceforge.stripes.controller.LifecycleStage;
import net.sourceforge.stripes.controller.StripesConstants;
import net.sourceforge.stripes.controller.StripesFilter;
import net.sourceforge.stripes.exception.StripesServletException;
import net.sourceforge.stripes.validation.BooleanTypeConverter;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * <p>
 * A Magnolia paragraph renderer that delegates to Stripes actions. Most of the code is just a cut and paste from
 * Stripes' {@link DispatcherServlet}, adapted to work withing magnolia by:
 * </p>
 * <ul>
 * <li>wrap the request in order to modify the request path and provide a custom requestDispatcher</li>
 * <li>wrap the response in order to provide a custom Writer</li>
 * <li>use a fake servlet instance/context to setup a Stripe context (there is no servlet here)</li>
 * <li>injiect any paragraph property as a parameter</li>
 * </ul>
 * <p>
 * <strong>Todo:</strong>
 * </p>
 * <ul>
 * <li>Handle multipart forms in request wrapper (should be easy to do)</li>
 * <li>A better way of handling multivalued properties in paragraph</li>
 * <li>Handle binary properties in paragraph</li>
 * </ul>
 * @author fgiust
 * @version $Id: $
 */
public class StripesParagraphRenderer implements ParagraphRenderer
{

    private Boolean alwaysInvokeValidate;

    /**
     * Logger.
     */
    private Logger log = LoggerFactory.getLogger(StripesParagraphRenderer.class);

    /**
     * <p>
     * Invokes the following instance level methods in order to coordinate the processing of requests:
     * </p>
     * <ul>
     * <li>{@link #resolveActionBean(ExecutionContext)}</li>
     * <li>{@link #resolveHandler(ExecutionContext)}</li>
     * <li>{@link #doBindingAndValidation(ExecutionContext)}</li>
     * <li>{@link #doCustomValidation(ExecutionContext)}</li>
     * <li>{@link #handleValidationErrors(ExecutionContext)}</li>
     * <li>{@link #invokeEventHandler(ExecutionContext)}</li>
     * </ul>
     * <p>
     * If any of the above methods return a {@link Resolution} the rest of the request processing is aborted and the
     * resolution is executed.
     * </p>
     * {@inheritDoc}
     */
    @SuppressWarnings("unchecked")
    public void render(Content content, Paragraph paragraph, Writer out) throws IOException
    {
        // @fgiust start custom magnolia setup

        WebContext webContext = (WebContext) MgnlContext.getInstance();
        HttpServletResponse response = new StripesParagraphResponseWrapper(webContext.getResponse(), out);

        final String templatePath = paragraph.getTemplatePath();
        Map<String, String[]> nodeDataMap = contentToMap(content);

        HttpServletRequest request;
        try
        {
            request = new StripesParagraphRequestWrapper(
                ((WebContext) MgnlContext.getInstance()).getRequest(),
                templatePath,
                nodeDataMap);

            ((StripesParagraphRequestWrapper) request).setLocale(MgnlContext.getLocale());
        }
        catch (StripesServletException e)
        {
            throw new RuntimeException(e);
        }

        // @todo fgiust force include instead of forwarding. Not needed anymore since we are wrapping requestDispatcher?
        request.setAttribute(StripesConstants.REQ_ATTR_INCLUDE_PATH, templatePath);

        PageContext pageContext = webContext.getPageContext();
        ServletContext servletContext = null;

        if (pageContext != null)
        {
            pageContext.getServletContext();
        }
        // @fgiust end custom magnolia setup

        // It sucks that we have to do this here (in the request cycle), but there doesn't
        // seem to be a good way to get at the Configuration from the Filter in init()
        doOneTimeConfiguration();

        log.debug("Dispatching request to URL: ", request.getRequestURI());

        try
        {
            final Configuration config = StripesFilter.getConfiguration();

            // First manufacture an ActionBeanContext
            final ActionBeanContext context = config
                .getActionBeanContextFactory()
                .getContextInstance(request, response);
            context.setServletContext(servletContext);

            // Then setup the ExecutionContext that we'll use to process this request
            final ExecutionContext ctx = new ExecutionContext();
            ctx.setInterceptors(config.getInterceptors(LifecycleStage.ActionBeanResolution));
            ctx.setLifecycleStage(LifecycleStage.ActionBeanResolution);
            ctx.setActionBeanContext(context);

            ActionBeanContext abc = ctx.getActionBeanContext();
            if (pageContext != null)
            {
                DispatcherHelper.setPageContext(pageContext);
            }

            // Resolve the ActionBean, and if an interceptor returns a resolution, bail now
            saveActionBean(request);
            Resolution resolution = DispatcherHelper.resolveActionBean(ctx);

            if (resolution == null)
            {
                resolution = DispatcherHelper.resolveHandler(ctx);

                if (resolution == null)
                {
                    // Then run binding and validation
                    resolution = DispatcherHelper.doBindingAndValidation(ctx, true);

                    if (resolution == null)
                    {
                        // Then continue on to custom validation
                        resolution = DispatcherHelper.doCustomValidation(ctx, alwaysInvokeValidate);

                        if (resolution == null)
                        {
                            // And then validation error handling
                            resolution = DispatcherHelper.handleValidationErrors(ctx);

                            if (resolution == null)
                            {
                                // And finally(ish) invoking of the event handler
                                resolution = DispatcherHelper.invokeEventHandler(ctx);

                                // If the event produced errors, fill them in
                                DispatcherHelper.fillInValidationErrors(ctx);
                            }
                        }
                    }
                }
            }

            // Whatever stage it came from, execute the resolution
            if (resolution != null)
            {
                DispatcherHelper.executeResolution(ctx, resolution);
            }
        }
        catch (RuntimeException re)
        {
            throw re;
        }
        catch (InvocationTargetException ite)
        {
            if (ite.getTargetException() instanceof RuntimeException)
            {
                throw (RuntimeException) ite.getTargetException();
            }
            else
            {
                throw new RuntimeException("ActionBean execution threw an exception.", ite.getTargetException());
            }
        }
        catch (Exception e)
        {
            throw new RuntimeException("Exception encountered processing request.", e);
        }
        finally
        {

            if (pageContext != null)
            {
                // *Don't* release the page context, if set it will be closed by the main magnolia template
                // JspFactory.getDefaultFactory().releasePageContext(pageContext);
                DispatcherHelper.setPageContext(null);
            }
            restoreActionBean(request);
        }
    }

    /**
     * @param content paragraph node
     * @return a map of Strings (converted nodedata)
     */
    @SuppressWarnings("unchecked")
    protected Map<String, String[]> contentToMap(Content content)
    {
        Collection<NodeData> paragraphsData = content.getNodeDataCollection();
        Map<String, String[]> nodeDataMap = new HashMap<String, String[]>();
        for (NodeData nodeData : paragraphsData)
        {
            String name = StringUtils.replaceChars(nodeData.getName(), "{}", "[]");
            String value = nodeData.getString();
            if (StringUtils.contains(name, "multiple"))
            {
                nodeDataMap.put(name, StringUtils.split(value, "\n"));
            }
            else
            {
                nodeDataMap.put(name, new String[]{value });
            }
        }
        return nodeDataMap;
    }

    /**
     * Performs a simple piece of one time configuration that requires access to the Configuration object delivered
     * through the Stripes Filter.
     */
    private void doOneTimeConfiguration()
    {
        if (alwaysInvokeValidate == null)
        {
            // Check to see if, in this application, validate() methods should always be run
            // even when validation errors already exist
            String callValidateWhenErrorsExist = StripesFilter
                .getConfiguration()
                .getBootstrapPropertyResolver()
                .getProperty(DispatcherServlet.RUN_CUSTOM_VALIDATION_WHEN_ERRORS);

            if (callValidateWhenErrorsExist != null)
            {
                BooleanTypeConverter c = new BooleanTypeConverter();
                this.alwaysInvokeValidate = c.convert(callValidateWhenErrorsExist, Boolean.class, null);
            }
            else
            {
                this.alwaysInvokeValidate = false; // Default behaviour
            }
        }
    }

    /**
     * Fetches, and lazily creates if required, a Stack in the request to store ActionBeans should the current request
     * involve forwards or includes to other ActionBeans.
     * @param request the current HttpServletRequest
     * @param create create a Stack if not already set
     * @return the Stack if present, or if creation is requested
     */
    @SuppressWarnings("unchecked")
    protected Stack getActionBeanStack(HttpServletRequest request, boolean create)
    {
        Stack stack = (Stack) request.getAttribute(StripesConstants.REQ_ATTR_ACTION_BEAN_STACK);
        if (stack == null && create)
        {
            stack = new Stack();
            request.setAttribute(StripesConstants.REQ_ATTR_ACTION_BEAN_STACK, stack);
        }

        return stack;
    }

    /**
     * Saves the current value of the 'actionBean' attribute in the request so that it can be restored at a later date
     * by calling {@link #restoreActionBean(HttpServletRequest)}. If no ActionBean is currently stored in the request,
     * nothing is changed.
     * @param request the current HttpServletRequest
     */
    @SuppressWarnings("unchecked")
    protected void saveActionBean(HttpServletRequest request)
    {
        if (request.getAttribute(StripesConstants.REQ_ATTR_ACTION_BEAN) != null)
        {
            Stack stack = getActionBeanStack(request, true);
            stack.push(request.getAttribute(StripesConstants.REQ_ATTR_ACTION_BEAN));
        }
    }

    /**
     * Restores the previous value of the 'actionBean' attribute in the request. If no ActionBeans have been saved using
     * {@link #saveActionBean(HttpServletRequest)} then this method has no effect.
     * @param request the current HttpServletRequest
     */
    @SuppressWarnings("unchecked")
    protected void restoreActionBean(HttpServletRequest request)
    {
        Stack stack = getActionBeanStack(request, false);
        if (stack != null && !stack.empty())
        {
            request.setAttribute(StripesConstants.REQ_ATTR_ACTION_BEAN, stack.pop());
        }
    }

}
